# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

from ..base import Config, Constant
from ..page import *
from ..language import *
from .page_iterator import *
from .parser import *


class BothIterator:

    def __init__(self, name = None, *, language = None):
        self.config = Config(name)
        self.name = self.config.name
        self.page_iterator = PageIterator(self.name)

        if language is None:
            self.language = Language.get_class(self.config.get_language(must = True))
        else:
            if language not in Constant.available_languages:
                raise ValueError(f"Language \"{language}\" is not supported.")
            else:
                self.language = Language.get_class(language)

    def __iter__(self):
        self.page_iterator.__iter__()
        return self

    def __next__(self):

        while True:
            page = next(self.page_iterator)
            item = Parser.page_to_class(page, language = self.language)

            if item is not None:
                return item



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.


from .wscraper_exception import *
from .constant import *
from .config import *
from .builder import *



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

from .language_base import *


class Japanese(LanguageBase):

    exclusive_title = [
        "メディア",
        "特別",
        "ノート",
        "利用者",
        "利用者‐会話",
        "Wikipedia",
        "Wikipedia‐ノート",
        "ファイル",
        "ファイル‐ノート",
        "MediaWiki",
        "MediaWiki‐ノート",
        "Template",
        "Template‐ノート",
        "Help",
        "Help‐ノート",
        "Category",
        "Category‐ノート",
        "Portal",
        "Portal‐ノート",
        "プロジェクト",
        "プロジェクト‐ノート",
        "モジュール",
        "モジュール‐ノート",
        "Gadget",
        "Gadget talk",
        "Gadget definition",
        "Gadget definition talk",
        "Topic",
    ]

    @classmethod
    def tag_to_surface(cls, re_tag_match):
        name = re_tag_match.group(1)
        body = re_tag_match.group(3)

        if name in ["div", "span", "sub", "sup"]:
            return body
        elif name == "code":
            if len(body) <= 16:
                return body
            else:
                return "コード"
        elif name == "math":
            return "数式"
        else:
            return ""

    @classmethod
    def bracket_to_surface(cls, re_bracket_match):
        items = re_bracket_match.group(1).split("|")
        top = items[0]

        if top == "仮リンク":
            return items[-1]
        elif top in ["IPA", "unicode", "lang"]:
            return items[-1]
        else:
            return ""

    @classmethod
    def link_to_surface(cls, re_link_match):
        entities = re_link_match.group(1).split("|")

        if len(entities) == 1:
            entity = entities[0]
            if entity.startswith("Category:"):
                return ""
            else:
                return entity
        elif len(entities) == 2:
            e0, e1 = entities
            if e0.startswith("ファイル:") or e0.startswith("File:"):
                return "ファイル"
            else:
                return e1
        else:
            return ""

    @classmethod
    def link_to_title(cls, re_link_match):
        entities = re_link_match.group(1).split("|")

        if len(entities) == 1:
            return entities[0]
        elif len(entities) == 2:
            e0, e1 = entities
            if e0.startswith("ファイル:") or e0.startswith("File:"):
                return None
            else:
                return e1
        else:
            return None

    @classmethod
    def document_to_sentences(cls, document):
        sentences = []

        while True:
            index = document.find("。")

            if index == -1:
                if len(document) > 0:
                    sentences.append(document.strip("\n"))
                break

            sentence = document[:index + 1]
            sentences.append(sentence.strip("\n"))
            document = document[index + 1:]

        return sentences



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

class Entry:

    def __init__(self, title, mediawiki):
        self.title = title
        self.mediawiki = mediawiki



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import re

from .language_base import *


class English(LanguageBase):

    exclusive_title = [
        "Media",
        "Special",
        "Talk",
        "User",
        "User talk",
        "Wikipedia",
        "Wikipedia talk",
        "File",
        "File talk",
        "MediaWiki",
        "MediaWiki talk",
        "Template",
        "Template talk",
        "Help",
        "Help talk",
        "Category",
        "Category talk",
        "Portal",
        "Portal talk",
        "Book",
        "Book talk",
        "Draft",
        "Draft talk",
        "Education Program",
        "Education Program talk",
        "TimedText",
        "TimedText talk",
        "Module",
        "Module talk",
        "Image"
    ]

    @classmethod
    def tag_to_surface(cls, re_tag_match):
        name = re_tag_match.group(1)
        body = re_tag_match.group(3)

        if name in ["div", "span", "sub", "sup"]:
            return body
        elif name == "code":
            if len(body) <= 16:
                return body
            else:
                return "code"
        elif name == "math":
            return "math"
        else:
            return ""

    @classmethod
    def bracket_to_surface(cls, re_bracket_match):
        items = re_bracket_match.group(1).split("|")
        top = items[0]

        if top == "仮リンク":
            return items[-1]
        elif top in ["IPA", "unicode", "lang"]:
            return items[-1]
        else:
            return ""

    @classmethod
    def link_to_surface(cls, re_link_match):
        entities = re_link_match.group(1).split("|")

        if len(entities) == 1:
            entity = entities[0]
            if entity.startswith("Category:"):
                return ""
            else:
                return entity
        elif len(entities) == 2:
            e0, e1 = entities
            if e0.startswith("File:"):
                return "File"
            else:
                return e1
        else:
            return ""

    @classmethod
    def document_to_sentences(cls, document):
        re_eos = re.compile(r"([.?!])\s+", re.DOTALL)

        sentences = []

        while True:
            match = re.search(re_eos, document)

            if match is None:
                if len(document) > 0:
                    sentences.append(document.strip("\n"))
                break

            sentence = document[:match.end(1)]
            sentences.append(sentence)
            document = document[match.end():]

        return sentences



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

from ..page import *
from .both_iterator import *


class RedirectionIterator:

    def __init__(self, name = None, *, language = None):
        self.both_iterator = BothIterator(name, language = language)

    def __iter__(self):
        self.both_iterator.__iter__()
        return self

    def __next__(self):

        while True:
            page = next(self.both_iterator)

            if page is not None and type(page) == Redirection:
                return page



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

from .wscraper_language_error import *


class Language:

    @classmethod
    def get_class(cls, name):
        if name == "japanese":
            from .japanese import Japanese
            return Japanese
        elif name == "english":
            from .english import English
            return English
        else:
            raise WScraperLanguageError(f"Language `{name}` is not implemented.")



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import re
import sys
import glob
import os

from ..utility import *
from .config import *
from .wscraper_exception import *


class Builder:

    @classmethod
    def command_import(cls, wikipedia_xml, wikipedia_name, *, move = False, copy = False, reset = False, language = None):
        Config.check_root_directory_exists()

        if not wikipedia_xml.endswith(".xml"):
            raise WScraperException("XML file required.")

        if move and copy:
            raise WScraperException("Both move and copy are True. Nothing or one should be True.")

        if wikipedia_name is None:
            wikipedia_name = os.path.basename(wikipedia_xml)[:-4]

        target_directory = os.path.join(Constant.wikipedia_directory, wikipedia_name)

        if os.path.isdir(target_directory):
            if not reset:
                sys.stdout.write("Split XML files are already exist. Skipping.\n")
                return
            else:
                sys.stdout.write("Split XML files are exist. Removing.\n")
                for path in glob.glob(os.path.join(target_directory, "*.xml")):
                    os.remove(path)
        else:
            os.makedirs(target_directory)

        sys.stdout.write(f"Split Wikipedia XML file `{wikipedia_xml}` to `{target_directory}`.\n")
        cls.split_to_xmls(wikipedia_xml, target_directory, chunk = Config.load_root_config()["page_chunk"])
        sys.stdout.write("Done!\n")

        if copy:
            sys.stdout.write(f"File is copying.\n")
            new_path = shutil.copy(wikipedia_xml, target_directory + ".xml")
            sys.stdout.write(f"File was copied to {new_path}.\n")
        elif move:
            new_path = shutil.move(wikipedia_xml, target_directory + ".xml")
            sys.stdout.write(f"File was moved {wikipedia_xml} to {new_path}.\n")

        Config.create_config(wikipedia_name, exist_ok = True)
        if language is not None:
            config = Config(wikipedia_name)
            config.set_language(language)
            config.save()
            sys.stdout.write(f"Language {language} is set to {wikipedia_name}.\n")

    @classmethod
    def split_to_xmls(cls, xml_path, output_directory, *, chunk = 10000):
        if not (type(chunk) == int and chunk >= Constant.min_page_chunk and chunk <= Constant.max_page_chunk):
            raise WScraperException("Config page_chunk must satisfy 100 <= chunk <= 10 ** 7.")

        re_start = re.compile(r"<page>")
        re_end = re.compile(r"</page>")

        flag = False
        body = []

        reader = open(xml_path, "r")
        writer = FileWriter(output_directory, "{:04d}.xml", chunk, prefix = "<wikipedia>\n", suffix = "</wikipedia>\n")

        progress = ProgressManager(writer)

        with reader, writer, progress:
            for line in reader:
                progress.update()

                if re_start.search(line):
                    flag = True

                if flag:
                    body.append(line)

                if re_end.search(line):
                    flag = False
                    writer.write("".join(body), count = 1)
                    body = []



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.


class LanguageBase:

    @classmethod
    def escape_decode(cls, string):
        return string.replace("&amp;", "&")

    @classmethod
    def extract_category(cls, link):
        category = "Category:"

        if link.startswith(category):
            return link[len(category):].rstrip("\u200e")
        else:
            return None



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import os
import glob

from ..base import *
from ..utility import *


class PageIterator:

    def __init__(self, name = None):
        self.config = Config(name)
        self.xml_directory = self.config.get_wikipedia_xml_directory(must = True)

    def __iter__(self):
        self.path_list = sorted(glob.glob(os.path.join(self.xml_directory, "*.xml")))
        self.n_path = len(self.path_list)
        self.i_path = 0
        self.page_list = []
        return self

    def __next__(self):
        if len(self.page_list) == 0:
            if len(self.path_list) == 0:
                raise StopIteration()

            path = self.path_list.pop(0)
            self.page_list = FileManager.load_xml(path).findall("page")
            self.i_path += 1

        return self.page_list.pop(0)

    def for_progress_manager(self):
        return [{ "name": "reader", "value": f"{self.i_path}/{self.n_path}" }]



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import sys
import argparse

from .base import *


help_string = """
wscraper

Wikipedia Scraper
    language model using Wikipedia

Command:
    wscraper initialize
        make root directory of wscraper and config file
    wscraper status
        describe status of current wikipedia corpus
    wscraper switch [name]
        change current wikipedia corpus to [name]
    wscraper import [xml path] [options...]
        read [xml path] and locate wikipedia resource directory
    wscraper set [parameters...]
        set parameters for current wikipedia corpus
    wscraper unset [parameters...]
        unset parameters for current wikipedia corpus
    wscraper rename [source] [target]
        rename wikipedia corpus from [source] to [target]
    wscraper root status
        check root configuration
    wscraper root set [parameters...]
        set root parameters
    wscraper root unset [parameters...]
        unset root parameters
    wscraper list
        list up names of wikipedia corpus resources
"""


def command():
    map_command_function = {
        "initialize": command_initialize,
        "status": command_status,
        "switch": command_switch,
        "import": command_import,
        "set": command_set,
        "unset": command_unset,
        "rename": command_rename,
        "delete": command_delete,
        "root": command_root,
        "list": command_list
    }

    if len(sys.argv) < 2 or sys.argv[1] in ["help", "-h", "--help"]:
        sys.stderr.write(f"{help_string}\n")
        sys.exit(1)

    command = sys.argv[1]

    if command not in map_command_function:
        sys.stderr.write(f"{help_string}\n")
        sys.exit(1)

    map_command_function[command](sys.argv[2:])


def command_initialize(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper initialize",
        description = "Command `wscraper initialize` creates primary files and directories."
    )

    args = parser.parse_args(args)

    Config.command_initialize()


def command_import(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper import",
        description = "Command `wscraper import` move a wikipedia.xml file to wscraper directory and split one to small files."
    )

    parser.add_argument("source", help = "source wikipedia.xml file")
    parser.add_argument("-n", "--name", help = "target directory name")
    parser.add_argument("-l", "--language", help = "language of wikipedia file")
    parser.add_argument("-m", "--move", action = "store_true", help = "move file to wscraper directory for backup")
    parser.add_argument("-c", "--copy", action = "store_true", help = "copy file to wscraper directory for backup")
    parser.add_argument("-r", "--reset", action = "store_true", help = "remove existing file")

    args = parser.parse_args(args)

    Builder.command_import(
        args.source, args.name,
        move = args.move, copy = args.copy, reset = args.reset, language = args.language
    )


def command_switch(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper switch",
        description = "Command `wscraper switch` switches the current wikipedia corpus."
    )

    parser.add_argument("name", help = "wikipedia name")

    args = parser.parse_args(args)

    Config.command_switch(args.name)


def command_rename(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper rename",
        description = "Command `wscraper rename` changes a name of wikipedia corpus."
    )

    parser.add_argument("source", help = "name from")
    parser.add_argument("target", help = "name to")

    args = parser.parse_args(args)

    Config.command_rename(args.source, args.target)


def command_delete(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper delete",
        description = "Command `wscraper delete` removes the wikipedia corpus."
    )

    parser.add_argument("target", help = "remove")

    args = parser.parse_args(args)

    Config.command_delete(args.target)


def command_status(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper status",
        description = "Command `wscraper status` shows a status of current wikipedia."
    )

    try:
        Config.command_status()
    except WScraperConfigError:
        sys.stderr.write("no wikipedia is set\n")
        sys.exit(1)


def command_set(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper set",
        description = "Command `wscraper set` sets value of parameter."
    )

    parser.add_argument("--language", help = "Wikipedia language")

    args = parser.parse_args(args)

    try:
        Config.command_set(language = args.language)
    except WScraperConfigError as e:
        sys.stderr.write(f"{e}\n")
        sys.exit(1)


def command_unset(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper unset",
        description = "Command `wscraper unset` unsets value of parameter."
    )

    parser.add_argument("--language", action = "store_true", help = "Wikipedia language")

    args = parser.parse_args(args)

    try:
        Config.command_unset(language = args.language)
    except WScraperConfigError as e:
        sys.stderr.write(f"{e}\n")
        sys.exit(1)


def command_list(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper list",
        description = "Command `wscraper list` shows available wikipedia names."
    )

    Config.command_list()


def command_root(args):
    if len(args) == 0:
        sys.stderr.write("Command `wscraper root` requires at least 1 argument.\n")
        sys.exit(1)

    sub_command = args[0]

    if sub_command == "status":
        Config.command_root_status()
    elif sub_command == "set":
        command_root_set(args[1:])
    elif sub_command == "unset":
        command_root_unset(args[1:])
    else:
        sys.stderr.write("Command `wscraper root` requires following string as a first argument. {status, set, unset}\n")
        sys.exit(1)


def command_root_set(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper root set",
        description = "Command `wscraper root set` sets default values of parameters."
    )

    parser.add_argument("--language", help = "language of wikipedia corpus")
    parser.add_argument("--page_chunk", type = int, help = "a unit of partition")

    args = parser.parse_args(args)

    Config.command_root_set(args.language, args.page_chunk)


def command_root_unset(args):
    parser = argparse.ArgumentParser(
        prog = "wscraper root unset",
        description = "Command `wscraper root unset` unsets default values of parameters."
    )

    parser.add_argument("--language", action = "store_true", help = "language of wikipedia corpus")

    args = parser.parse_args(args)

    Config.command_root_unset(args.language)



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import json
import xml.etree.ElementTree


class FileManager:

    @classmethod
    def load_xml(cls, path):
        with open(path, "r") as f:
            parser = xml.etree.ElementTree.XMLParser()

            for line in f:
                parser.feed(line)

            return parser.close()

    @classmethod
    def load_json(cls, path):
        with open(path, "r") as f:
            return json.load(f)

    @classmethod
    def save_json(cls, path, data):
        with open(path, "w") as f:
            json.dump(data, f, indent = 2)
            f.write("\n")



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import sys
import time


class ProgressManager:

    class Console:
        def __init__(self):
            self.last = -1

        def output(self, string, *, force = False):
            now = time.time()

            if force or now - self.last >= 1 / 60:
                self.last = now
                sys.stderr.write(string)
                sys.stderr.flush()

    class Notebook:
        def __init__(self):
            self.last = -1

        def output(self, string, *, force = False):
            now = time.time()

            if force or now - self.last >= 1:
                self.last = now
                import IPython.display as ds
                ds.clear_output()
                sys.stderr.write(string)
                sys.stderr.flush()

    def __init__(self, * instances):
        if "ipykernel" in sys.modules:
            self.output_type = self.Notebook()
        else:
            self.output_type = self.Console()

        self.instances = list(instances)
        self.count = 0

    def get_globals(self):
        return globals()

    def add(self, instance):
        self.instances.append(instance)

    def add_list(self, instances):
        self.instances.extend(instances)

    def add_count(self):
        self.count += 1

    def start(self):
        self.time = time.time()

    def finish(self):
        self.update(count_up = False, force = True, newline = True)

    def update(self, *, count_up = True, force = False, newline = False):
        now = time.time()

        interval = int(now - self.time)

        if count_up:
            self.add_count()

        second = interval % 60
        minute = interval % 3600 // 60
        hour = interval // 3600

        output_time = f"{hour}:{minute:02d}:{second:02d}"
        output_count = f"{self.count} count"

        output_base = f"\033[2K\033[G[progress] ({output_time}, {output_count})"

        if len(output_base) == 0:
            output_result = output_base
        else:
            output_others = [f"{x['name']}: {x['value']}" for inst in self.instances for x in inst.for_progress_manager()]
            output_result = f"{output_base}, {', '.join(output_others)}"

        if newline:
            output_result += "\n"

        self.output_type.output(output_result, force = force)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.finish()



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

class Redirection:

    def __init__(self, source, target):
        self.source = source
        self.target = target



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.


import re
import os
import sys
import shutil
import glob

from .constant import *
from .wscraper_exception import *
from ..utility import *


class WScraperConfigError(WScraperException):
    pass


class Config:

    template_root_config = {
        "language": None,
        "page_chunk": 10000
    }

    template_config = {
        "language": None
    }

    template_root_status = {
        "current": None
    }

    @classmethod
    def check_name_format(cls, name):
        return re.search(Constant.name_format, name) is not None

    @classmethod
    def check_root_directory_exists(cls):
        if not os.path.isdir(Constant.root_directory):
            message = f"Root directory {Constant.root_directory} does not exist."
            message += " Did you run `wscraper initialize`?"
            raise WScraperConfigError(message)

        if not os.path.isfile(Constant.root_config):
            raise WScraperConfigError(f"Root config file {Constant.root_config} does not exist.")

        if not os.path.isfile(Constant.root_status):
            raise WScraperConfigError(f"Root status file {Constant.root_status} does not exist.")

        if not os.path.isdir(Constant.wikipedia_directory):
            raise WScraperConfigError(f"Wikipedia directory {Constant.wikipedia_directory} does not exist.")

    @classmethod
    def command_root_status(cls):
        cls.check_root_directory_exists()

        config = cls.load_root_config()
        sys.stdout.write("\nRoot Status\n\n")
        sys.stdout.write(f"language: {config['language'] or '[not set]'}\n")
        sys.stdout.write(f"page_chunk: {config['page_chunk']}\n")
        sys.stdout.write("\n")

    @classmethod
    def command_root_set(cls, language, page_chunk):
        cls.check_root_directory_exists()

        config = cls.load_root_config()

        if language is not None:
            if language not in Constant.available_languages:
                raise WScraperConfigError(f"language `{language}` is not supported.")

            config["language"] = language

        if page_chunk is not None:
            if not (page_chunk >= Constant.min_page_chunk and page_chunk <= Constant.max_page_chunk):
                raise WScraperConfigError(f"page_chunk should satisfy {Constant.min_page_chunk} <= page_chunk <= {Constant.max_page_chunk}.")

            config["page_chunk"] = page_chunk

        FileManager.save_json(Constant.root_config, config)

    @classmethod
    def command_root_unset(cls, language):
        cls.check_root_directory_exists()

        config = cls.load_root_config()

        if language:
            config["language"] = None

        FileManager.save_json(Constant.root_config, config)

    @classmethod
    def command_initialize(cls):
        if os.path.isdir(Constant.root_directory):
            sys.stdout.write(f"Root directory {Constant.root_directory} already exists.\n")

            if os.path.isfile(Constant.root_config):
                sys.stdout.write(f"Root config {Constant.root_config} already exists.\n")
            else:
                FileManager.save_json(Constant.root_config, cls.template_root_config)
                sys.stdout.write(f"Root config {Constant.root_config} was created.\n")

            if os.path.isfile(Constant.root_status):
                sys.stdout.write(f"Root status {Constant.root_status} already exists.\n")
            else:
                FileManager.save_json(Constant.root_status, cls.template_root_status)
                sys.stdout.write(f"Root status {Constant.root_status} was created.\n")

            if os.path.isdir(Constant.wikipedia_directory):
                sys.stdout.write(f"Wikipedia directory {Constant.wikipedia_directory} already exists.\n")
            else:
                os.makedirs(Constant.wikipedia_directory)
                sys.stdout.write(f"Wikipedia directory {Constant.wikipedia_directory} was created.\n")

        else:
            os.makedirs(Constant.root_directory)
            sys.stdout.write(f"Root directory {Constant.root_directory} was created.\n")
            FileManager.save_json(Constant.root_config, cls.template_root_config)
            sys.stdout.write(f"Root config {Constant.root_config} was creaetd.\n")
            FileManager.save_json(Constant.root_status, cls.template_root_status)
            sys.stdout.write(f"Root status {Constant.root_status} was created.\n")
            os.makedirs(Constant.wikipedia_directory)
            sys.stdout.write(f"Wikipedia directory {Constant.wikipedia_directory} was created.\n")

    @classmethod
    def command_switch(cls, name):
        cls.check_root_directory_exists()

        if not os.path.isdir(os.path.join(Constant.wikipedia_directory, name)):
            raise WScraperConfigError(f"Wikipedia name {name} does not exist.")

        path = os.path.join(Constant.root_directory, "status.yml")
        status = FileManager.load_json(path)
        status["current"] = name
        FileManager.save_json(path, status)
        sys.stdout.write(f"Wikipedia switched to {name}.\n")

    @classmethod
    def command_set(cls, language):
        cls.check_root_directory_exists()

        config = Config()

        if language is not None:
            config.set_language(language)

        config.save()

    @classmethod
    def command_unset(cls, language):
        cls.check_root_directory_exists()

        config = Config()

        if language:
            config.unset_language()

        config.save()

    @classmethod
    def command_status(cls):
        cls.check_root_directory_exists()

        Config().print_status()

    @classmethod
    def command_rename(cls, source, target):
        cls.check_root_directory_exists()

        if not cls.check_name_format(target):
            raise WScraperConfigError(f"illegal name format: \"{target}\"")

        s = os.path.join(Constant.wikipedia_directory, source)
        sc = os.path.join(Constant.wikipedia_directory, f"{source}.config.json")
        sx = os.path.join(Constant.wikipedia_directory, f"{source}.xml")

        if not os.path.isdir(s) or not os.path.isfile(sc):
            raise WScraperConfigError(f"no such wikipedia corpus: \"{source}\"")

        t = os.path.join(Constant.wikipedia_directory, target)
        tc = os.path.join(Constant.wikipedia_directory, f"{target}.config.json")
        tx = os.path.join(Constant.wikipedia_directory, f"{target}.xml")

        if os.path.exists(t) or os.path.exists(tc):
            raise WScraperConfigError(f"Wikipedia corpus {target} already exists.")

        shutil.move(s, t)
        shutil.move(sc, tc)

        if os.path.isfile(sx):
            shutil.move(sx, tx)

        sys.stdout.write(f"rename: {source} -> {target}\n")

    @classmethod
    def command_delete(cls, target):
        cls.check_root_directory_exists()

        t = os.path.join(Constant.wikipedia_directory, target)
        tc = os.path.join(Constant.wikipedia_directory, f"{target}.config.json")
        tx = os.path.join(Constant.wikipedia_directory, f"{target}.xml")

        if os.path.isdir(t) and os.path.isfile(tc):
            shutil.rmtree(t)
            os.remove(tc)
            if os.path.isfile(tx):
                os.remove(tx)
            sys.stdout.write(f"delete: {target}\n")
        else:
            raise WScraperConfigError(f"no such wikipedia corpus: \"{target}\"")

    @classmethod
    def command_list(cls):
        sys.stdout.write("\n")

        list_wikipedia = cls.list_wikipedia()

        if len(list_wikipedia) == 0:
            sys.stdout.write("Available Wikipedia is nothing.\n")
        else:
            sys.stdout.write("Available Wikipedia:\n")
            for item in list_wikipedia:
                sys.stdout.write(f"  - {item}\n")

        sys.stdout.write("\n")

    @classmethod
    def list_wikipedia(cls):
        items = []

        for path in glob.glob(os.path.join(Constant.wikipedia_directory, "*")):
            if os.path.isdir(path):
                items.append(os.path.basename(path))

        return items

    @classmethod
    def load_root_config(cls):
        return FileManager.load_json(Constant.root_config)

    @classmethod
    def load_root_status(cls):
        return FileManager.load_json(Constant.root_status)

    @classmethod
    def load_config_by_name(cls, name):
        path = os.path.join(Constant.wikipedia_directory, f"{name}.config.json")

        if not os.path.isfile(path):
            raise WScraperConfigError(f"Wikipedia name {name} does not exist.")

        return FileManager.load_json(path)

    @classmethod
    def validate_name(cls, name):
        if re.search(Constant.name_format, name) is None:
            raise WScraperConfigError(f"invalid name `{name}`.")

    @classmethod
    def create_config(cls, name, *, exist_ok = False):
        path = os.path.join(Constant.wikipedia_directory, f"{name}.config.json")

        if os.path.isfile(path):
            if not exist_ok:
                raise WScraperConfigError(f"Wikipedia name {name} already exists.")
        else:
            cls.validate_name(name)
            FileManager.save_json(path, cls.template_config)

    def __init__(self, name = None):
        self.check_root_directory_exists()

        if name is None:
            status = FileManager.load_json(Constant.root_status)
            name = status["current"]
            if name is None:
                raise WScraperConfigError("Current name is not set.")

        self.name = name
        self.config = self.load_config_by_name(name)
        self.root_config = self.load_root_config()
        self.root_status = self.load_root_status()
        self.path = os.path.join(Constant.wikipedia_directory, f"{self.name}.config.json")

    def save(self):
        FileManager.save_json(self.path, self.config)

    def print_status(self):
        print("\ncurrent:", self.name)

        print()

        language = self.config["language"]

        if language is None:
            print("language [defalt]:", self.root_config["language"] or "[not set]")
        else:
            print("language:", language)

        print()

    def get_parameter(self, key, *, must = False):
        data = self.config

        for k in key.split("."):
            if data is not None:
                data = data.get(k)

        if data is None and key in self.template_root_config:
            data = self.root_config[key]

        if data is None and must:
            raise WScraperConfigError(f"Key {key} must exist but None.")

        return data

    def set_language(self, language):
        if language not in Constant.available_languages:
            raise WScraperConfigError(f"Language `{language}` is not supported.")

        self.config["language"] = language
        self.save()

    def unset_language(self):
        self.config["language"] = None
        self.save()

    def get_language(self, must = False):
        return self.get_parameter("language", must = must)

    def get_wikipedia_xml(self, *, must = True):
        wikipedia_xml = os.path.join(Constant.wikipedia_directory + f"{self.name}.xml")

        if must and not os.path.isfile(wikipedia_xml):
            raise WScraperConfigError(f"No such file {wikipedia_xml}.")

        return wikipedia_xml

    def get_wikipedia_xml_directory(self, *, must = True):
        return os.path.join(Constant.wikipedia_directory, f"{self.name}")



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import os
from setuptools import setup, find_packages


def get_readme():
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md")) as f:
        return f.read()


setup(
    name = "wscraper",
    version = "0.1.0",
    license = "MIT License",
    author = "T.Furukawa",
    author_email = "tfurukawa.mail@gmail.com",
    url = "https://github.com/tfull/wikipedia_scraper",
    packages = find_packages(),
    keywords = ["Wikipedia"],
    description = "Scraping documents from a dump XML file of Wikipedia.",
    long_description = get_readme(),
    long_description_content_type = "text/markdown",
    entry_points = {
        "console_scripts": [
            "wscraper = wscraper.console:command"
        ]
    }
)



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.


import os


class Constant:

    root_directory = os.environ.get("WSCRAPER_ROOT") or os.path.join(os.environ["HOME"], ".wscraper")

    root_config = os.path.join(root_directory, "config.yml")
    root_status = os.path.join(root_directory, "status.yml")

    wikipedia_directory = os.path.join(root_directory, "wikipedia")

    name_format = r"^[a-zA-Z0-9_.\-]+$"

    available_languages = [
        "japanese",
        "english"
    ]

    min_page_chunk = 100
    max_page_chunk = 10 ** 7
    min_worker = 1
    max_worker = 64



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

import re

from ..utility import *
from ..page import *


RE_COMMENT    = re.compile(r"<!--.*?-->", re.DOTALL)
RE_NOWIKI     = re.compile(r"<nowiki>.*?</nowiki>", re.DOTALL)
RE_BRACKET_1  = re.compile(r"\{[^\{]*?\}")
RE_BRACKET_2  = re.compile(r"\{\{([^\{]*?)\}\}")
RE_REF_SINGLE = re.compile(r"<ref[^>]*/>")
RE_REF_PAIR   = re.compile(r"<ref[^>]*>.*?</ref>", re.DOTALL)
RE_PRIME      = re.compile(r"\'{2,}")
RE_TAG        = re.compile(r"<([a-zA-Z]+)(.*?)?>(.*?)</\1>")
RE_LINK       = re.compile(r"\[\[([^\[]*?)\]\]")
RE_SQ         = re.compile(r"\[[^\[]*?\]")
RE_CHAPTER    = re.compile(r"={2,}(.*?)={2,}")
RE_LIST       = re.compile(r"\* ")


class Parser:

    @classmethod
    def page_to_class(cls, page, *, language = None):
        if language is None:
            raise ValueError("language: None")

        title = page.find("title").text

        for rm in language.exclusive_title:
            if title.startswith(rm + ":"):
                return None

        redirect = page.find("redirect")

        if redirect is not None:
            target = redirect.attrib["title"]

            for rm in language.exclusive_title:
                if target.startswith(rm + ":"):
                    return None

            return Redirection(source = title, target = target)

        revision = page.find("revision")
        text = revision.find("text").text

        if text is None:
            return None

        text = "".join(filter(lambda c: ord(c) < 0x10000, text))

        return Entry(title = title, mediawiki = text)

    @classmethod
    def interpret_mediawiki(cls, mediawiki, *, language = None):
        if language is None:
            raise ValueError("language: None")

        text = mediawiki

        text = re.sub(RE_COMMENT, "", text)
        text = re.sub(RE_NOWIKI, "", text)

        while True:
            match = re.search(RE_BRACKET_2, text)

            if match is None:
                break

            text = text[: match.start()] + language.bracket_to_surface(match) + text[match.end() :]

        text = re.sub(RE_BRACKET_1, "", text)
        text = re.sub(RE_REF_SINGLE, "", text)
        text = re.sub(RE_REF_PAIR, "", text)
        text = re.sub(RE_PRIME, "", text)

        text = text.replace("__TOC__", "").replace("&amp;", "&")

        while True:
            match = re.search(RE_TAG, text)

            if match is None:
                break

            text = text[: match.start()] + language.tag_to_surface(match) + text[match.end() :]

        while True:
            match = re.search(RE_LINK, text)

            if match is None:
                break

            text = text[: match.start()] + language.link_to_surface(match) + text[match.end() :]

        return text

    @classmethod
    def extract_links(cls, mediawiki, *, language = None):
        if language is None:
            ValueError("language: None")

        text = mediawiki
        links = []

        while True:
            match = re.search(RE_LINK, text)

            if match is None:
                break

            text = text[: match.start()] + language.link_to_surface(match) + text[match.end() :]
            title = language.link_to_title(match)

            if title is not None:
                links.append(title)

        return links

    @classmethod
    def extract_categories(cls, mediawiki, *, language = None):
        links = cls.extract_links(mediawiki, language = language)
        categories = []

        for link in links:
            category = language.extract_category(link)

            if category is not None:
                categories.append(category)

        return categories

    @classmethod
    def to_plain_text(cls, mediawiki, *, language = None):
        if language is None:
            raise ValueError("language: None")

        text = cls.interpret_mediawiki(mediawiki, language = language)
        lines = []

        for line in text.split("\n"):
            if len(line) == 0:
                continue

            if line.startswith("* "):
                lines.append(line[2:])
            else:
                match = re.match(RE_CHAPTER, line)

                if match:
                    pass
                else:
                    lines.append(line)

        return "\n".join(lines)

    @classmethod
    def to_paragraphs(cls, mediawiki, *, language = None):
        if language is None:
            raise ValueError("language: None")

        text = cls.interpret_mediawiki(mediawiki, language = language)
        lines = text.split("\n")

        paragraphs = []
        lines_stock = []
        previous = "*"

        for line in lines:
            match = re.match(RE_CHAPTER, line)

            if match:
                paragraphs.append((previous, "\n".join(lines_stock)))
                previous = match.group(1)
                lines_stock = []
            elif len(line) > 0:
                if line.startswith("* "):
                    lines_stock.append(line[2:])
                else:
                    lines_stock.append(line)

        paragraphs.append((previous, "\n".join(lines_stock)))

        return [(title, document) for title, document in paragraphs if len(document) > 0]

    @classmethod
    def to_sentences(cls, mediawiki, *, language = None):
        if language is None:
            raise ValueError("language: None")

        plain_text = cls.to_plain_text(mediawiki, language = language)

        return language.document_to_sentences(plain_text)



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.


from .parser import *
from .page_iterator import *
from .both_iterator import *
from .entry_iterator import *
from .redirection_iterator import *



# Copyright (c) 2021 T.Furukawa
# This software is released under the MIT License, see LICENSE.

from collections import OrderedDict


class FileWriter:

    def __init__(self, output_directory, path_format, threshold, *, prefix = None, suffix = None):
        self.output_directory = output_directory
        self.path_format = path_format
        self.threshold = threshold
        self.count = 0
        self.total_count = 0
        self.file_count = 0
        self.file_object = None
        self.prefix = prefix
        self.suffix = suffix

    def write(self, string, count = 1):
        if self.file_object is None:
            self.file_count += 1
            self.file_object = open(f"{self.output_directory}/{self.path_format.format(self.file_count)}", "w")
            if self.prefix is not None:
                self.file_object.write(self.prefix)

        self.file_object.write(string)
        self.count += count
        self.total_count += count

        if self.count >= self.threshold:
            if self.suffix is not None:
                self.file_object.write(self.suffix)
            self.file_object.close()
            self.file_object = None
            self.count = 0

    def finish(self):
        if self.file_object is not None:
            if self.suffix is not None:
                self.file_object.write(self.suffix)
            self.file_object.close()
            self.file_object = None

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.finish()

    def for_progress_manager(self):
        return [{ "name": "writer", "value": self.total_count }, { "name": "page", "value": self.file_count }]
